import numpy as np
import rdkit.Chem.rdmolfiles as rdmolfiles

from aamutils.utils import mol_to_graph
from aamutils.algorithm.ilp import expand_partial_aam_balanced


def test_e2e():
    mol_G = rdmolfiles.MolFromSmiles("C=[O:1].[O:2]")
    mol_H = rdmolfiles.MolFromSmiles("C([O:2])[O:1]")
    exp_X = np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]])
    G = mol_to_graph(mol_G)
    H = mol_to_graph(mol_H)
    X, status, v = expand_partial_aam_balanced(G, H)
    np.testing.assert_array_equal(X, exp_X)
    assert 4 == v
    assert "Optimal" == status


def test_example1():
    mol_G = rdmolfiles.MolFromSmiles("[C:1][C:2][C:3][Cl].[N]")
    mol_H = rdmolfiles.MolFromSmiles("[C:1][C:2][C:3][N].[Cl]")
    exp_X = np.zeros((5, 5))
    exp_X[0, 0] = 1
    exp_X[1, 1] = 1
    exp_X[2, 2] = 1
    exp_X[3, 4] = 1
    exp_X[4, 3] = 1
    G = mol_to_graph(mol_G)
    H = mol_to_graph(mol_H)
    X, status, v = expand_partial_aam_balanced(G, H)
    assert 4 == v
    assert "Optimal" == status
    np.testing.assert_array_equal(X, exp_X)


def test_example2_m1():
    mol_G = rdmolfiles.MolFromSmiles("[C:1][C:2]=[C:3][C:4]")
    mol_H = rdmolfiles.MolFromSmiles("[C:1].[C:2]=[C:4][C:3]")
    exp_X = np.zeros((4, 4))
    exp_X[0, 0] = 1
    exp_X[1, 1] = 1
    exp_X[2, 3] = 1
    exp_X[3, 2] = 1
    G = mol_to_graph(mol_G)
    H = mol_to_graph(mol_H)
    X, status, v = expand_partial_aam_balanced(G, H)
    assert "Optimal" == status
    assert 6 == v
    np.testing.assert_array_equal(X, exp_X)


def test_example2_m2():
    mol_G = rdmolfiles.MolFromSmiles("CC=CC")
    mol_H = rdmolfiles.MolFromSmiles("C.C=CC")
    exp_X = np.eye(4)
    G = mol_to_graph(mol_G)
    H = mol_to_graph(mol_H)
    X, status, v = expand_partial_aam_balanced(G, H)
    assert "Optimal" == status
    assert 2 == v
    np.testing.assert_array_equal(X, exp_X)


def test_edge_diff_constraint():
    G_smiles = "CCCC"
    H_smiles = "CCCC"
    mol_G = rdmolfiles.MolFromSmiles(G_smiles)
    mol_H = rdmolfiles.MolFromSmiles(H_smiles)
    G = mol_to_graph(mol_G)
    H = mol_to_graph(mol_H)
    X, status, v = expand_partial_aam_balanced(G, H)
    assert v == 0
    assert "Optimal" == status



import collections
import numpy as np
import networkx as nx
import rdkit.Chem as Chem
import rdkit.Chem.rdmolfiles as rdmolfiles
import rdkit.Chem.rdDepictor as rdDepictor


def mol_to_graph(mol: Chem.rdchem.Mol) -> nx.Graph:
    bond_order_map = {
        "SINGLE": 1,
        "DOUBLE": 2,
        "TRIPLE": 3,
        "QUADRUPLE": 4,
        "AROMATIC": 1.5,
    }
    g = nx.Graph()
    for atom in mol.GetAtoms():
        aam = atom.GetAtomMapNum()
        formal_charge = atom.GetFormalCharge()  # Get the formal charge of the atom
        g.add_node(
            atom.GetIdx(), symbol=atom.GetSymbol(), aam=aam, formal_charge=formal_charge
        )  # Store formal charge
    for bond in mol.GetBonds():
        bond_type = str(bond.GetBondType()).split(".")[-1]
        bond_order = 1
        if bond_type in bond_order_map.keys():
            bond_order = bond_order_map[bond_type]
        g.add_edge(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(), bond=bond_order)
    return g


def smiles_to_graph(
    smiles: str, sanitize: bool = True
) -> nx.Graph | tuple[nx.Graph, nx.Graph]:
    if ">>" in smiles:
        smiles_token = smiles.split(">>")
        g = mol_to_graph(rdmolfiles.MolFromSmiles(smiles_token[0], sanitize=sanitize))
        h = mol_to_graph(rdmolfiles.MolFromSmiles(smiles_token[1], sanitize=sanitize))
        return g, h
    else:
        return mol_to_graph(rdmolfiles.MolFromSmiles(smiles, sanitize=sanitize))


def graph_to_mol(
    G: nx.Graph, symbol_key="symbol", aam_key="aam", bond_type_key="bond"
) -> Chem.rdchem.Mol:
    bond_order_map = {
        1: Chem.rdchem.BondType.SINGLE,
        2: Chem.rdchem.BondType.DOUBLE,
        3: Chem.rdchem.BondType.TRIPLE,
        4: Chem.rdchem.BondType.QUADRUPLE,
        1.5: Chem.rdchem.BondType.AROMATIC,
    }
    rw_mol = Chem.rdchem.RWMol()
    idx_map = {}
    for n, d in G.nodes(data=True):
        idx = rw_mol.AddAtom(Chem.rdchem.Atom(d[symbol_key]))
        idx_map[n] = idx
        if aam_key in d.keys() and d[aam_key] >= 0:
            rw_mol.GetAtomWithIdx(idx).SetAtomMapNum(d[aam_key])
        if "formal_charge" in d:  # Set the formal charge for the atom
            rw_mol.GetAtomWithIdx(idx).SetFormalCharge(d["formal_charge"])
    for n1, n2, d in G.edges(data=True):
        idx1 = idx_map[n1]
        idx2 = idx_map[n2]
        rw_mol.AddBond(idx1, idx2, bond_order_map[d[bond_type_key]])
    return rw_mol.GetMol()


def get_beta_map(G, H, aam_key="aam"):
    node2aam_G = collections.defaultdict(
        lambda: -1, {n: d[aam_key] for (n, d) in G.nodes(data=True) if d[aam_key] > 0}
    )
    aam2node_H = collections.defaultdict(
        lambda: -1, {d[aam_key]: n for (n, d) in H.nodes(data=True) if d[aam_key] > 0}
    )
    beta_map = []
    for n_G, aam_G in node2aam_G.items():
        if aam_G in aam2node_H and aam2node_H[aam_G] > -1:
            beta_map.append((n_G, aam2node_H[aam_G], aam_G))
    return beta_map


def set_aam(G, H, M, beta_map=None, aam_key="aam"):
    if beta_map is None:
        beta_map = get_beta_map(G, H)
    used_atom_numbers = [aam for _, _, aam in beta_map]

    aam_G = collections.defaultdict(lambda: -1)
    for bi, _, aam in beta_map:
        aam_G[bi] = aam

    next_aam_nr = 1
    for n, d in G.nodes(data=True):
        if aam_G[n] > -1:
            aam_nr = aam_G[n]
        else:
            while next_aam_nr in used_atom_numbers:
                next_aam_nr += 1
            aam_nr = next_aam_nr
            used_atom_numbers.append(aam_nr)
        d[aam_key] = int(aam_nr)
        aam_G[n] = aam_nr

    aam_G = np.array([v for _, v in sorted(aam_G.items(), key=lambda x: x[0])])
    for (_, d), aam in zip(H.nodes(data=True), np.dot(aam_G.T, M)):
        d[aam_key] = int(aam)


def is_valid_rxn_smiles(smiles):
    smiles_token = smiles.split(">>")
    if len(smiles_token) != 2:
        return False
    mol1 = rdmolfiles.MolFromSmiles(smiles_token[0])
    if mol1 is None:
        return False
    mol2 = rdmolfiles.MolFromSmiles(smiles_token[0])
    if mol2 is None:
        return False
    return True


def print_graph(graph):
    print(
        "Graph Nodes: {}".format(
            " ".join(
                [
                    "[{}]{}:{}".format(
                        n[0], n[1].get("symbol", None), n[1].get("aam", None)
                    )
                    for n in graph.nodes(data=True)
                ]
            )
        )
    )
    print(
        "Graph Edges: {}".format(
            " ".join(
                [
                    "[{}]-[{}]:{}".format(n[0], n[1], n[2]["bond"])
                    for n in graph.edges(data=True)
                ]
            )
        )
    )


def its2mol(its: nx.Graph, aam_key="aam", bond_key="bond") -> Chem.rdchem.Mol:
    _its = its.copy()
    for n in _its.nodes:
        _its.nodes[n][aam_key] = n
    for u, v in _its.edges():
        _its[u][v][bond_key] = 1
    return graph_to_mol(_its)


def plot_its(
    its, ax, bond_key="bond", aam_key="aam", symbol_key="symbol", use_mol_coords=True
):
    bond_char = {None: "∅", 1: "—", 2: "=", 3: "≡"}
    mol = its2mol(its, aam_key=aam_key, bond_key=bond_key)

    if use_mol_coords:
        positions = {}
        conformer = rdDepictor.Compute2DCoords(mol)
        for i, atom in enumerate(mol.GetAtoms()):
            aam = atom.GetAtomMapNum()
            apos = mol.GetConformer(conformer).GetAtomPosition(i)
            positions[aam] = [apos.x, apos.y]
    else:
        positions = nx.spring_layout(its)

    ax.axis("equal")
    ax.axis("off")

    nx.draw_networkx_edges(its, positions, edge_color="#000000", ax=ax)
    nx.draw_networkx_nodes(its, positions, node_color="#FFFFFF", node_size=500, ax=ax)

    labels = {n: "{}:{}".format(d[symbol_key], n) for n, d in its.nodes(data=True)}
    edge_labels = {}
    for u, v, d in its.edges(data=True):
        bc1 = d[bond_key][0]
        bc2 = d[bond_key][1]
        if bc1 == bc2:
            continue
        if bc1 in bond_char.keys():
            bc1 = bond_char[bc1]
        if bc2 in bond_char.keys():
            bc2 = bond_char[bc2]
        edge_labels[(u, v)] = "({},{})".format(bc1, bc2)

    nx.draw_networkx_labels(its, positions, labels=labels, ax=ax)
    nx.draw_networkx_edge_labels(its, positions, edge_labels=edge_labels, ax=ax)



from aamutils.mapping import Mapper
from aamutils.algorithm.aaming import get_its, get_rc
from aamutils.utils import smiles_to_graph


def test_mapper_with_expected_rc():
    # Diels-Alder reaction example
    smiles = "C1=CC=CC1.C1=CC=CC1>>C1=CC2CC1C3C2C=CC3"
    mapper = Mapper()
    aam_smiles = mapper.get_aam(smiles, expected_rc=(6, 6))
    assert isinstance(aam_smiles, str)
    g, h = smiles_to_graph(aam_smiles)
    its = get_its(g, h)
    rc = get_rc(its)
    assert len(rc.nodes) == 6
    assert len(rc.edges) == 6


def test_mapper_infeasable():
    smiles = "C>>C"
    mapper = Mapper()
    aam_smiles = mapper.get_aam(smiles, expected_rc=(2, 2))
    assert aam_smiles is None



import networkx as nx
import rdkit.Chem.rdmolfiles as rdmolfiles
from aamutils.algorithm.ilp import expand_partial_aam_balanced
from aamutils.utils import smiles_to_graph, graph_to_mol, set_aam


def extend_aam_from_graph(G: nx.Graph, H: nx.Graph) -> str:
    """
    Extends atom-atom mappings (AAM) from two input graphs,
    G (reactants) and H (products),by solving the AAM problem using an
    Integer Linear Programming (ILP) approach and generating
    a reaction SMILES (RSMI) string.

    Parameters:
    - G (nx.Graph): Graph representing the reactants.
    Nodes should contain necessary attributes for AAM.
    - H (nx.Graph): Graph representing the products.
    Nodes should contain necessary attributes for AAM.

    Returns:
    - str: A reaction SMILES string (RSMI) in the format 'reactant>>product'
    with extended atom mappings.
    """
    # Solve the partial AAM problem using ILP and retrieve the mapping matrix
    M, _, _ = expand_partial_aam_balanced(G, H)

    # Apply the AAM matrix to the graphs
    set_aam(G, H, M)

    # Convert the modified graphs back to RDKit molecules
    r_mol = graph_to_mol(G)
    p_mol = graph_to_mol(H)

    # Generate reaction SMILES string from the RDKit molecules
    result_smiles = "{}>>{}".format(
        rdmolfiles.MolToSmiles(
            r_mol, canonical=True, kekuleSmiles=False, allHsExplicit=True
        ),
        rdmolfiles.MolToSmiles(
            p_mol, canonical=True, kekuleSmiles=False, allHsExplicit=True
        ),
    )
    return result_smiles


def extend_aam_from_rsmi(rsmi: str) -> str:
    """
    Extends atom-atom mappings (AAM) from a reaction SMILES (RSMI) string,
    and returns the resulting reaction SMILES string.

    Parameters:
    - rsmi (str): A reaction SMILES string in the format 'reactant>>product'.

    Returns:
    - str: A reaction SMILES string with extended atom mappings.
    """
    G, H = smiles_to_graph(rsmi)
    return extend_aam_from_graph(G, H)



import argparse

from aamutils.cmd.expanding import configure_parser as configure_expanding_parser


def setup_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="AAM Utils",
        description="A collection of atom-atom-mapping utility functions.",
    )
    subparsers = parser.add_subparsers(dest="command")

    configure_expanding_parser(subparsers)

    return parser



import time
import traceback
import copy
import json
import collections
import random
import argparse
import networkx as nx
import numpy as np

import rdkit.Chem.rdmolfiles as rdmolfiles

from aamutils.algorithm.ilp import expand_partial_aam_balanced
from aamutils.algorithm.aaming import get_its, get_rc
from aamutils.utils import mol_to_graph, set_aam


def load_data():
    with open("data/test_dataset.json", "r") as f:
        _data = json.load(f)

    data = []
    rc_size_hist = collections.defaultdict(lambda: 0)
    for i, entry in enumerate(_data):
        if entry["equivalent"] is False:
            continue
        try:
            _entry = copy.deepcopy(entry)
            smiles = _entry["local_mapper"]
            smiles = smiles.split(">>")
            g_mol = rdmolfiles.MolFromSmiles(smiles[0])
            h_mol = rdmolfiles.MolFromSmiles(smiles[1])
            G = mol_to_graph(g_mol)
            H = mol_to_graph(h_mol)
            if len(G.nodes) != len(H.nodes):
                print("Skip because unbalanced")
                continue
            ITS = get_its(G, H)
            RC = get_rc(ITS)
            _entry["G"] = G
            _entry["H"] = H
            _entry["ITS"] = ITS
            _entry["RC"] = RC
            rc_size_hist[len(RC.nodes)] += 1
            data.append(_entry)
        except Exception:
            print("Error at index {}".format(i))

    print("{} of {} are equivalent".format(len(data), len(_data)))
    return data


def run(n, remove_mode, remove_ratio, seed=None):
    if seed is not None:
        random.seed(seed)
    data = load_data()
    testcase_cnt = 0
    success_cnt = 0
    start_time = time.time()
    for i, entry in enumerate(data):
        try:
            remove_cnt = 0

            nodes = []
            if remove_mode == "keep_rc":
                rc_nodes = list(entry["RC"].nodes)
                nodes = list(entry["ITS"].nodes)
                nodes = [n for n in nodes if n not in rc_nodes]
            elif remove_mode == "rc":
                nodes = list(entry["RC"].nodes)
            else:
                raise ValueError()
            samples = random.sample(nodes, int(len(nodes) * remove_ratio))

            for rand_n in samples:
                G_idx, H_idx = nx.get_node_attributes(entry["ITS"], "idx_map")[rand_n]
                remove_cnt += 1
                entry["G"].nodes[G_idx]["aam"] = 0
                entry["H"].nodes[H_idx]["aam"] = 0

            M, status, value = expand_partial_aam_balanced(entry["G"], entry["H"])

            set_aam(entry["G"], entry["H"], M)
            ITS = get_its(entry["G"], entry["H"])
            RC = get_rc(ITS)

            success = nx.is_isomorphic(
                entry["ITS"],
                ITS,
                node_match=lambda n1, n2: n1["symbol"] == n2["symbol"],
                edge_match=lambda e1, e2: e1["bond"] == e2["bond"],
            )

            testcase_cnt += 1
            if success:
                success_cnt += 1

            print(
                (
                    "[{:>6}|{:>4}] {} {:>2} {} | Removed {} ids. "
                    + "RC Nodes: {}->{} Edges: {}->{} | "
                    + "ETA: {}"
                ).format(
                    entry["R-id"],
                    testcase_cnt,
                    status,
                    int(value),
                    "SUCC" if success else "FAIL",
                    remove_cnt,
                    len(entry["RC"].nodes),
                    len(RC.nodes),
                    len(entry["RC"].edges),
                    len(RC.edges),
                    time.strftime(
                        "%H:%M:%S",
                        time.gmtime(
                            int(time.time() - start_time)
                            * ((np.min([len(data), n]) - testcase_cnt) / testcase_cnt)
                        ),
                    ),
                )
            )

            if testcase_cnt == n:
                break
        except Exception as e:
            print("[{}] Error: {}".format(entry["R-id"], e))
            traceback.print_exc()

    print(
        ("Expanding was successful in {:.2%} ({} out of {} testcases).").format(
            success_cnt / testcase_cnt, success_cnt, testcase_cnt
        )
    )


if __name__ == "__main__":
    default_rm_mode = "rc"
    default_rm_ratio = 0.5
    default_n = 1000
    default_seed = None

    parser = argparse.ArgumentParser(
        prog="benchmarks.py",
        description="Script to run the benchmarks from the paper.",
    )
    parser.add_argument(
        "-n",
        type=int,
        default=default_n,
        help="The number of test cases. Default: {}".format(default_n),
    )
    parser.add_argument(
        "--remove-mode",
        choices=["rc", "keep_rc"],
        default=default_rm_mode,
        help=(
            "This argument selects from which atoms atom-numbers are removed "
            + "for test case generation. "
            + "(1) 'rc': remove AAM from reaction center "
            + "(use --remove-ratio to set the amount of removed atom-numbers), "
            + "(2) 'keep_rc': remove AAM from all other atoms except the "
            + "reaction center. Default: {}"
        ).format(default_rm_mode),
    )
    parser.add_argument(
        "--remove-ratio",
        type=float,
        default=default_rm_ratio,
        help=(
            "The ratio ([0, 1]) of atom-numbers to remove. "
            + "The ratio specifies the upper bound: a ratio of 0.5 will "
            + "remove 1 atom-number in a reaction center with 3 atoms."
            + " Default: {}"
        ).format(default_rm_ratio),
    )
    parser.add_argument(
        "--seed",
        default=default_seed,
        help=(
            "The seed used to initialize random number generation. " + "Default: {}"
        ).format(default_seed),
    )

    args = parser.parse_args()
    run(args.n, args.remove_mode, args.remove_ratio, args.seed)



import rdkit.Chem.rdmolfiles as rdmolfiles

from aamutils.utils import mol_to_graph
from aamutils.algorithm.aaming import get_its, get_rc


def test_its():
    g_smiles = "[I:1][I:9].[cH:6]1[cH:7][cH:2][cH:3][cH:4][cH:5]1"
    h_smiles = "[I:1][c:2]1[cH:3][cH:4][cH:5][cH:6][cH:7]1.[H+].[I-:9]"
    g_mol = rdmolfiles.MolFromSmiles(g_smiles)
    h_mol = rdmolfiles.MolFromSmiles(h_smiles)
    G = mol_to_graph(g_mol)
    H = mol_to_graph(h_mol)
    ITS = get_its(G, H)
    RC = get_rc(ITS)
    assert 3 == len(RC.nodes)
    assert all(i in RC.nodes for i in [1, 2, 9])


def test_partial_aam_with_unmapped_H():
    g_smiles = "[CH2:1]=[O:2].[CH3:3][NH2:4].[H][H]"
    h_smiles = "[CH3:3][NH:4][CH3:1].[OH2:2]"
    g_mol = rdmolfiles.MolFromSmiles(g_smiles)
    h_mol = rdmolfiles.MolFromSmiles(h_smiles)
    G = mol_to_graph(g_mol)
    H = mol_to_graph(h_mol)
    ITS = get_its(G, H)
    RC = get_rc(ITS)
    assert 4 == len(ITS.nodes)
    assert 3 == len(RC.nodes)
    assert all([i in ITS.nodes for i in [1, 2, 3, 4]])
    assert all([i in RC.nodes for i in [1, 2, 4]])
    assert 2 == len(RC.edges)
    assert (2, None) == RC[1][2]["bond"]
    assert (None, 1) == RC[1][4]["bond"]



import os
import json
import logging
import rdkit.Chem.rdmolfiles as rdmolfiles
import rdkit.Chem.rdmolops as rdmolops

from aamutils.utils import mol_to_graph, set_aam, graph_to_mol, is_valid_rxn_smiles
from aamutils.algorithm.ilp import expand_partial_aam_balanced
from aamutils.algorithm.aaming import get_its, get_rc, is_rc_valid

logger = logging.getLogger("aamutils")


def run(args):
    outputfile = args.o
    inputfile = args.inputfile
    if outputfile is None:
        in_path, in_ext = os.path.splitext(inputfile)
        outputfile = "{}_extended.json".format(in_path)
    rxn_smiles = []
    with open(inputfile, "r") as f:
        rxn_smiles = [line.strip() for line in f if is_valid_rxn_smiles(line)]

    results = []
    success_cnt = 0
    for smiles in rxn_smiles:
        result = {"input": smiles}
        try:
            r_smiles, p_smiles = smiles.split(">>")
            r_mol = rdmolfiles.MolFromSmiles(r_smiles)
            p_mol = rdmolfiles.MolFromSmiles(p_smiles)
            r_mol = rdmolops.AddHs(r_mol, 1)
            p_mol = rdmolops.AddHs(p_mol, 1)
            G = mol_to_graph(r_mol)
            H = mol_to_graph(p_mol)

            M, status, e_diff = expand_partial_aam_balanced(G, H)

            set_aam(G, H, M)

            r_mol = graph_to_mol(G)
            p_mol = graph_to_mol(H)

            its = get_its(G, H)
            rc = get_rc(its)
            rc_invalid = not is_rc_valid(rc)
            result_smiles = "{}>>{}".format(
                rdmolfiles.MolToSmiles(r_mol), rdmolfiles.MolToSmiles(p_mol)
            )
            result["expanded_aam"] = result_smiles
            result["ilp_status"] = status
            result["optimization_result"] = e_diff
            result["invalid_reaction_center"] = rc_invalid
            result["reaction_edges"] = len(rc.nodes)
            success_cnt += 1
        except Exception as e:
            result["error"] = str(e)
        finally:
            results.append(result)

    with open(outputfile, "w") as f:
        json.dump(results, f, indent=4)
    logger.info(
        "Expanded {} atom-atom-maps (failed: {})".format(
            success_cnt, len(rxn_smiles) - success_cnt
        )
    )
    logger.info("The result is written to file '{}'.".format(outputfile))


def configure_parser(subparsers):
    parser = subparsers.add_parser(
        "expand", help="Utility to expand a partial atom-atom-maps."
    )
    parser.add_argument(
        "inputfile",
        help=(
            "A file containing reaction smiles with "
            + "partial atom-atom-maps. Reactions should be separated by newlines."
        ),
    )
    parser.add_argument("-o", default=None, help="Path to output file.")

    parser.set_defaults(func=run)



import unittest
from aamutils.utils import smiles_to_graph
from aamutils.aam_expand import extend_aam_from_graph, extend_aam_from_rsmi


class TestExtendAAMFromGraph(unittest.TestCase):
    def test_extend_aam_from_graph(self):
        """
        Tests the extend_aam_from_graph function to ensure it correctly extends
        atom-atom mappings and returns the expected reaction SMILES.
        """
        rsmi = "CC[CH2:3][Cl:1].[N:2]>>CC[CH2:3][N:2].[Cl:1]"
        G, H = smiles_to_graph(rsmi)
        result_smiles = extend_aam_from_graph(G, H)
        expected = (
            "[Cl:1][CH2:3][CH2:5][CH3:4].[NH3:2]>>[ClH:1].[NH2:2][CH2:3][CH2:5][CH3:4]"
        )
        self.assertEqual(result_smiles, expected)

    def test_extend_aam_from_smiles(self):
        """
        Tests the extend_aam_from_graph function to ensure it correctly extends
        atom-atom mappings and returns the expected reaction SMILES.
        """
        rsmi = "CC[CH2:3][Cl:1].[N:2]>>CC[CH2:3][N:2].[Cl:1]"

        result_smiles = extend_aam_from_rsmi(rsmi)
        expected = (
            "[Cl:1][CH2:3][CH2:5][CH3:4].[NH3:2]>>[ClH:1].[NH2:2][CH2:3][CH2:5][CH3:4]"
        )
        self.assertEqual(result_smiles, expected)


# Run the unittest
if __name__ == "__main__":
    unittest.main()



import collections
import networkx as nx


def _add_its_nodes(ITS, G, H, eta, symbol_key):
    eta_G, eta_G_inv, eta_H, eta_H_inv = eta[0], eta[1], eta[2], eta[3]
    for n, d in G.nodes(data=True):
        n_ITS = eta_G[n]
        n_H = eta_H_inv[n_ITS]
        if n_ITS is not None and n_H is not None:
            ITS.add_node(n_ITS, symbol=d[symbol_key], idx_map=(n, n_H))
    for n, d in H.nodes(data=True):
        n_ITS = eta_H[n]
        n_G = eta_G_inv[n_ITS]
        if n_ITS is not None and n_G is not None and n_ITS not in ITS.nodes:
            ITS.add_node(n_ITS, symbol=d[symbol_key], idx_map=(n_G, n))


def _add_its_edges(ITS, G, H, eta, bond_key):
    eta_G, eta_G_inv, eta_H, eta_H_inv = eta[0], eta[1], eta[2], eta[3]
    for n1, n2, d in G.edges(data=True):
        if n1 > n2:
            continue
        e_G = d[bond_key]
        n_ITS1 = eta_G[n1]
        n_ITS2 = eta_G[n2]
        n_H1 = eta_H_inv[n_ITS1]
        n_H2 = eta_H_inv[n_ITS2]
        e_H = None
        if H.has_edge(n_H1, n_H2):
            e_H = H[n_H1][n_H2][bond_key]
        if not ITS.has_edge(n_ITS1, n_ITS2) and n_ITS1 > 0 and n_ITS2 > 0:
            ITS.add_edge(n_ITS1, n_ITS2, bond=(e_G, e_H))

    for n1, n2, d in H.edges(data=True):
        if n1 > n2:
            continue
        e_H = d[bond_key]
        n_ITS1 = eta_H[n1]
        n_ITS2 = eta_H[n2]
        n_G1 = eta_G_inv[n_ITS1]
        n_G2 = eta_G_inv[n_ITS2]
        if n_G1 is None or n_G2 is None:
            continue
        if not G.has_edge(n_G1, n_G2) and n_ITS1 > 0 and n_ITS2 > 0:
            ITS.add_edge(n_ITS1, n_ITS2, bond=(None, e_H))


def get_its(
    G: nx.Graph, H: nx.Graph, aam_key="aam", symbol_key="symbol", bond_key="bond"
) -> nx.Graph:
    """Get the ITS graph of reaction G \u2192 H.

    :param G: Educt molecular graph.
    :param H: Product molecular graph.
    :param aam_key: (optional) The node label that encodes atom indices on the
        molecular graph.
    :param symbol_key: (optional) The node label that encodes atom symbols on
        the molecular graph.
    :param bond_key: (optional) The edge label that encodes bond order on the
        molecular graph.

    :returns: Returns the ITS graph.
    """
    eta_G = collections.defaultdict(lambda: None)
    eta_G_inv = collections.defaultdict(lambda: None)
    eta_H = collections.defaultdict(lambda: None)
    eta_H_inv = collections.defaultdict(lambda: None)
    eta = (eta_G, eta_G_inv, eta_H, eta_H_inv)

    for n, d in G.nodes(data=True):
        if aam_key in d.keys() and d[aam_key] >= 0:
            eta_G[n] = d[aam_key]
            eta_G_inv[d[aam_key]] = n
    for n, d in H.nodes(data=True):
        if aam_key in d.keys() and d[aam_key] >= 0:
            eta_H[n] = d[aam_key]
            eta_H_inv[d[aam_key]] = n

    ITS = nx.Graph()
    _add_its_nodes(ITS, G, H, eta, symbol_key)
    _add_its_edges(ITS, G, H, eta, bond_key)

    return ITS


def get_rc(ITS: nx.Graph, symbol_key="symbol", bond_key="bond") -> nx.Graph:
    """Get the reaction center (RC) graph from an ITS graph.

    :param ITS: The ITS graph to get the RC from.
    :param symbol_key: (optional) The node label that encodes atom symbols on
        the molecular graph.
    :param bond_key: (optional) The edge label that encodes bond order on the
        molecular graph.

    :returns: Returns the reaction center (RC) graph.
    """
    rc = nx.Graph()
    for n1, n2, d in ITS.edges(data=True):
        edge_label = d[bond_key]
        if edge_label[0] != edge_label[1]:
            rc.add_node(n1, **{symbol_key: ITS.nodes[n1][symbol_key]})
            rc.add_node(n2, **{symbol_key: ITS.nodes[n2][symbol_key]})
            rc.add_edge(n1, n2, **{bond_key: edge_label})
    return rc


def is_rc_valid(RC: nx.Graph, symbol_key="symbol") -> bool:
    """Checks if a reaction center (RC) graph is valid. A valid RC does not
    mean it is chemically correct. This function only checks the necessary
    structural requirements for a correct RC. A chemically correct RC is always
    valid but a valid RC is not necessarily chemically correct.

    :param symbol_key: (optional) The node label that encodes atom symbols on
        the molecular graph.

    :returns: Returns ``True`` if the RC graph is valid, ``False`` otherwise.
    """
    for n, d in RC.degree():
        if d < 2:
            return False, "Atom {} {} has degree {}.".format(
                n, RC.nodes[n][symbol_key], d
            )
    return True



import rdkit.Chem.rdmolops as rdmolops
import rdkit.Chem.rdmolfiles as rdmolfiles

from aamutils.utils import graph_to_mol, mol_to_graph, set_aam
from aamutils.algorithm.ilp import expand_partial_aam_balanced


class Mapper:
    def __init__(self, include_H=False):
        self.include_H = include_H

    def _map_rxn(self, rxn_smiles: str, expected_rc):
        g_smiles, h_smiles = rxn_smiles.split(">>")
        g_mol = rdmolfiles.MolFromSmiles(g_smiles)
        h_mol = rdmolfiles.MolFromSmiles(h_smiles)
        if self.include_H:
            g_mol = rdmolops.AddHs(g_mol, 0)
            h_mol = rdmolops.AddHs(h_mol, 0)
        G = mol_to_graph(g_mol)
        H = mol_to_graph(h_mol)

        M, status, e_diff = expand_partial_aam_balanced(G, H, expected_rc=expected_rc)
        if status != "Optimal":
            return None

        set_aam(G, H, M)

        g_mol = graph_to_mol(G)
        h_mol = graph_to_mol(H)

        aam_smiles = "{}>>{}".format(
            rdmolfiles.MolToSmiles(g_mol), rdmolfiles.MolToSmiles(h_mol)
        )
        return aam_smiles

    def get_aam(self, smiles: str | list[str], expected_rc=None) -> str | list[str]:
        return_list = True
        if isinstance(smiles, str):
            smiles = [smiles]
            return_list = False
        if expected_rc is None:
            expected_rc = [None for _ in range(len(smiles))]
        elif not isinstance(expected_rc, list):
            expected_rc = [expected_rc]
        if len(smiles) != len(expected_rc):
            raise ValueError(
                "Argument min_bonds must be of same length ({}) as smiles.".format(
                    len(smiles)
                )
            )
        aam_smiles = []
        for smiles, exp_rc in zip(smiles, expected_rc):
            aam_smiles.append(self._map_rxn(smiles, expected_rc=exp_rc))
        if not return_list:
            return aam_smiles[0]
        return aam_smiles



import numpy as np
import pulp as lp
import networkx as nx

from aamutils.utils import get_beta_map


def _bijection_constraint(problem, X):
    m = int(np.sqrt(len(X.keys())))

    for j in range(m):
        problem += (
            lp.lpSum([X[i, j] for i in range(m)]) == 1,
            "bijection_col_{}".format(j),
        )
    for i in range(m):
        problem += (
            lp.lpSum([X[i, j] for j in range(m)]) == 1,
            "bijection_row_{}".format(i),
        )


def _beta_constraint(problem, X, beta_map):
    for bi, bj, _ in beta_map:
        problem += (X[bi, bj] == 1, "beta_constraint1_{}-{}".format(bi, bj))


def _edge_diff_constraint(problem, X, D, A_G, A_H):
    m = int(np.sqrt(len(D.keys())))
    for i in range(m):
        for j in range(m):
            problem += D[i, j] == lp.lpSum(
                [X[i, k] * A_H[k, j] for k in range(m)]
            ) - lp.lpSum([A_G[i, k] * X[k, j] for k in range(m)])


def _atom_type_constraint(problem, X, G, H, symbol_key="symbol"):
    m = int(np.sqrt(len(X.keys())))
    assert m == len(G.nodes)
    assert m == len(H.nodes)
    for i in range(m):
        for j in range(m):
            g_sym = G.nodes(data=True)[i][symbol_key]
            h_sym = H.nodes(data=True)[j][symbol_key]
            assert isinstance(g_sym, str)
            assert isinstance(h_sym, str)
            if g_sym != h_sym:
                problem += (X[i, j] == 0, "atom_type_constraint_{}-{}".format(i, j))


def _indicator_constraint(problem, D, G, S, k):
    m = int(np.sqrt(len(D.keys())))
    assert m == int(np.sqrt(len(G.keys())))
    assert m == int(np.sqrt(len(S.keys())))

    for i in range(m):
        for j in range(m):
            problem += (
                D[i, j] <= k * G[i, j],
                "g_indicator_constraint1_{}-{}".format(i, j),
            )
            problem += (
                D[i, j] >= -k + G[i, j] * (k + 1),
                "g_indicator_constraint2_{}-{}".format(i, j),
            )
            problem += (
                -D[i, j] <= k * S[i, j],
                "s_indicator_constraint1_{}-{}".format(i, j),
            )
            problem += (
                -D[i, j] >= -k + S[i, j] * (k + 1),
                "s_indicator_constraint2_{}-{}".format(i, j),
            )


def _expected_rc_atoms_constraint(problem, G, S, expected_rc_atoms):
    m = int(np.sqrt(len(G.keys())))
    assert m == int(np.sqrt(len(S.keys())))

    lp_io = lp.LpVariable.dicts("io", [i for i in range(m)], cat=lp.LpInteger)
    lp_rc = lp.LpVariable.dicts("rc", [i for i in range(m)], cat=lp.LpBinary)
    for i in range(m):
        problem += (
            lp_io[i] == lp.lpSum([G[i, j] + S[i, j] for j in range(m)]),
            "idx_io_sum_{}".format(i),
        )
        problem += (lp_io[i] <= m * lp_rc[i], "is_rc_indicator1_{}".format(i))
        problem += (lp_io[i] >= lp_rc[i], "is_rc_indicator_{}".format(i))

    problem += (
        expected_rc_atoms == lp.lpSum([lp_rc[i] for i in range(m)]),
        "rc_atom_constraint",
    )


def expand_partial_aam_balanced(
    G: nx.Graph,
    H: nx.Graph,
    beta_map: None | list[tuple[int, int, int]] = None,
    bond_key="bond",
    expected_rc: None | tuple[None | int, None | int] = None,
    time_limit: None | int = None,
) -> tuple[np.ndarray, str, float]:
    """Function to extend the partial atom-atom map beta of the balanced
    reaction G \u2192 H to a full atom-atom map based on the Minimal Chemical
    Distance (minimize the number of changing bonds).

    :param G: Educt molecular graph.
    :param H: Product molecular graph.
    :param beta_map: (optional) A list of predefined atom-atom mappings. If not
        specified the existing mapping in G and H will be used.
    :param bond_key: (optional) The edge label in G and H which encodes the
        bond order.
    :param expected_rc: (optional) Configuration for the expected reaction
        center. The first parameter in the tuple is the number of atoms and
        the second parameter is the number of bonds. The atoms and bonds are
        only constraint if the value is not None. It is possible to constraint
        only atoms or only bonds. Constraining the expected reaction center can
        help to get correct AAM for reactions where the Minimal Chemical
        Distance is unsuitable, e.g. for a Diels-Alder reaction you can set
        expected_rc to (6, 6).
    :param time_limit: (optional) Set a timeout for the solver.

    :returns: A 3-tuple. The first element is the mapping matrix X. The second
        element is the status string of the ILP solver (e.g. 'Optimal') and the
        third is the ILP solution value.
    """

    A_G = nx.adjacency_matrix(G, weight=bond_key).todense()
    A_H = nx.adjacency_matrix(H, weight=bond_key).todense()
    k = np.max([np.max(A_G), np.max(A_H)])
    m = len(G.nodes)

    if m != len(H.nodes):
        raise ValueError(
            (
                "Reaction is not balanced. " + "{} reactant atoms and {} product atoms."
            ).format(len(G.nodes), len(H.nodes))
        )

    if beta_map is None:
        beta_map = get_beta_map(G, H)

    # Mapping matrix X
    lp_X = lp.LpVariable.dicts(
        "X",
        [(i, j) for i in range(m) for j in range(m)],
        cat=lp.LpBinary,
    )

    # Difference matrix D
    lp_D = lp.LpVariable.dicts("D", [(i, j) for i in range(m) for j in range(m)])

    # Indicator matrix G and S
    lp_G = lp.LpVariable.dicts(
        "G",
        [(i, j) for i in range(m) for j in range(m)],
        cat=lp.LpBinary,
    )
    lp_S = lp.LpVariable.dicts(
        "S",
        [(i, j) for i in range(m) for j in range(m)],
        cat=lp.LpBinary,
    )

    # Non-zero counter function for D
    lp_f = lp.LpVariable("f", 0, None, lp.LpInteger)

    problem = lp.LpProblem("AAM", lp.LpMinimize)

    problem += (lp_f, "objective")
    problem += (
        lp_f == lp.lpSum([lp_G[i, j] + lp_S[i, j] for i in range(m) for j in range(m)]),
        "objective",
    )
    if expected_rc is not None:
        if not isinstance(expected_rc, tuple) or len(expected_rc) != 2:
            raise ValueError("Argument expected_rc must be a tuple of size 2.")
        expected_rc_atoms, expected_rc_bonds = expected_rc
        if expected_rc_bonds is not None:
            problem += (lp_f == 2 * expected_rc_bonds, "rc_bond_constraint")
        if expected_rc_atoms is not None:
            _expected_rc_atoms_constraint(problem, lp_G, lp_S, expected_rc_atoms)

    _bijection_constraint(problem, lp_X)
    _beta_constraint(problem, lp_X, beta_map)
    _atom_type_constraint(problem, lp_X, G, H)
    _edge_diff_constraint(problem, lp_X, lp_D, A_G, A_H)
    _indicator_constraint(problem, lp_D, lp_G, lp_S, k)

    status = problem.solve(lp.PULP_CBC_CMD(logPath=r"solver.log", timeLimit=time_limit))

    np_X = np.zeros([m, m], dtype=np.int32)
    for (i, j), v in lp_X.items():
        np_X[i, j] = int(v.value())

    return np_X, lp.LpStatus[status], lp_f.value()
